import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sphinx-astrorefs",
    version="0.9",
    author="Jo Bovy",
    author_email="bovy@astro.utoronto.ca",
    description="Astro-style references in Sphinx documents",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    packages=["sphinx_astrorefs"],
    package_data={"sphinx_astrorefs":["_static/sphinx_astrorefs.css"]},
    install_requires=["sphinx","sphinxcontrib-bibtex==2.1.4",
                      "pybtex","latexcodec"]
)



# pybtex_astro.py: Astro labeling and reference formatting
import re
from collections import Counter
import string
from pybtex.style.formatting.unsrt import Style as UnsrtStyle
from pybtex.style.formatting import toplevel
from pybtex.style.labels import BaseLabelStyle
from pybtex.style.sorting.author_year_title \
    import SortingStyle as AuthorSortingStyle
from pybtex.richtext import Symbol, Text
from pybtex.plugin import register_plugin
from pybtex.style.template import (
    field, first_of, href, join, names, optional, optional_field, sentence,
    tag, together, words, node, FieldIsMissing
)
import latexcodec
from sphinx.util import logging
logger= logging.getLogger(__name__)
_arxiv_url= None

def decode_specialchars(input):
    return input.replace('{', '').replace('}', '').encode().decode('latex')

# labels in astro format: author year, author & author year for two authors
# author et al. year for more than two
class AstroLabelStyle(BaseLabelStyle):
    def format_labels(self, sorted_entries):
        all_labels= []
        for entry in sorted_entries:
            if len(entry.persons['author']) == 1:
                out= '{} {}'.format(entry.persons['author'][0].last_names[0],
                                    entry.fields['year'])
            elif len(entry.persons['author']) == 2:
                out= '{} & {} {}'.format(entry.persons['author'][0]\
                                             .last_names[0],
                                         entry.persons['author'][1]\
                                              .last_names[0],
                                         entry.fields['year'])
            else:
                out= '{} et al. {}'.format(entry.persons['author'][0]\
                                               .last_names[0],
                                          entry.fields['year'])
            all_labels.append(decode_specialchars(out))
        # Deal with duplicates, assuming at most 26 duplicates
        dups= [item for item, count in Counter(all_labels).items()
               if count > 1]
        for dup in dups:
            idxs= [ii for ii,x in enumerate(all_labels) if x == dup]
            for idx,lett in zip(idxs,string.ascii_lowercase):
                last_digit= re.match('.+([0-9])[^0-9]*$',all_labels[idx])
                all_labels[idx]= all_labels[idx][:last_digit.start(1)+1]\
                    +lett+all_labels[idx][last_digit.start(1)+1:]
                sorted_entries[idx].fields['year']= \
                    sorted_entries[idx].fields['year']+lett
        # Yield output
        for entry, label in zip(sorted_entries,all_labels):
            yield label

def dashify(text):
    dash_re = re.compile(r'-+')
    return Text(Symbol('ndash')).join(text.split(dash_re))

def format_first_and_last_name(person):
    if len(person.first_names) > 0:
        return '{} {}.'.format(decode_specialchars(person.last_names[0]),
                               str(person.first_names[0])[0])
    else: # collaboration or similar
        return '{}'.format(decode_specialchars(person.last_names[0]))

@node
def astro_names(children, context, role, **kwargs):
    """Return formatted names."""
    assert not children
    try:
        persons= context['entry'].persons[role]
    except KeyError:
        raise FieldIsMissing(role, context['entry'])
    if len(persons) > 5:
        out= ''
        for ii in range(5):
            out+= '{}, '.format(format_first_and_last_name(persons[ii]))
        return '{}et al.'.format(out)
    elif len(persons) > 2:
        out= ''
        for ii in range(len(persons)-1):
            out+= '{}, '.format(format_first_and_last_name(persons[ii]))
        return '{}& {}'.format(out,
                                format_first_and_last_name(persons[-1]))
    elif len(persons) == 2:
        return '{} & {}'.format(format_first_and_last_name(persons[0]),
                                format_first_and_last_name(persons[1]))
    else:
        return format_first_and_last_name(persons[0])

class AstroStyle(UnsrtStyle):
    default_label_style = AstroLabelStyle
    default_sorting_style= AuthorSortingStyle

    def format_names(self, role, as_sentence=True):
        formatted_names= astro_names(role,sep=', ',
                                     sep2 = ' & ',last_sep=', & ')
        if as_sentence:
            return sentence [formatted_names]
        else:
            return formatted_names

    def format_author(self,e,as_sentence=True):
        authors= self.format_names('author', as_sentence=False)
        return sentence[authors]

    def format_journal(self,e):
        if 'doi' not in e.fields:
            return field('journal')
        else:
            return href [
                join [
                    'https://doi.org/',
                    field('doi', raw=True)
                    ],
                field('journal')
                ]

    def format_btitle(self,e, which_field, as_sentence=True):
        formatted_title = tag('em') [ field(which_field) ]
        if 'doi' not in e.fields and 'adsurl' in e.fields:
            url  = field('adsurl',raw=True)
        elif 'doi' in e.fields:
            url = join [
                'https://doi.org/',
                field('doi', raw=True)
                ]
        else:
            url = None
        if url is None and as_sentence:
            return sentence [formatted_title ]
        elif url is None:
            return formatted_title
        elif as_sentence:
            return sentence [
                href [
                    url,
                    formatted_title
                    ]
                ]
        else:
            return href [
                url,
                formatted_title
                ]

    def format_volume(self,e):
        if 'adsurl' not in e.fields:
            return field('volume')
        else:
            return href [ field('adsurl',raw=True),
                          field('volume')]

    def format_pages(self,e):
        if 'eprint' not in e.fields:
            return field('pages',apply_func=dashify)
        else:
            return href [
                join [
                    f'https://{_arxiv_url}/abs/',
                    field('eprint', raw=True)
                    ],
                field('pages',apply_func=dashify)
                ]

    def format_publisher(self,e):
        if 'doi' not in e.fields:
            return field('publisher')
        else:
            return href [
                join [
                    'https://doi.org/',
                    field('doi', raw=True)
                    ],
                field('publisher')
                ]

    def format_publisher_address(self,e):
        if 'adsurl' not in e.fields:
            return field('address')
        else:
            return href [ 
                field('adsurl',raw=True),
                field('address')
                ]
        
    def format_volume_and_series(self, e, as_sentence=True):
        if as_sentence:
            return sentence [
                join [
                    optional [ 
                        field('series') ,
                        ' ',
                        tag('strong') [ field('volume') ] 
                    ]
                ]
            ]
        else:
            return join [
                optional [ 
                    field('series') ,
                    ' ',
                    tag('strong') [ field('volume') ]
                ]
            ]
        
    def get_article_template(self, e):
        if 'volume' not in e.fields:
            journal_and_volume = tag('em') [self.format_journal(e)]
        else:
            journal_and_volume = join [
                tag('em') [self.format_journal(e)],' ',
                tag('strong') [self.format_volume(e)]
                ]
        template = toplevel [
            self.format_author(e),
            self.format_title(e, 'title'),
            sentence [
                journal_and_volume,
                join [
                    optional [self.format_pages(e)],
                    ' (',field('year'),')']
                ]
        ]
        return template

    def get_book_template(self, e):
        template = toplevel [
            self.format_author_or_editor(e),
            self.format_btitle(e, 'title'),
            self.format_volume_and_series(e),
            sentence [
                join [
                    self.format_publisher(e),
                    ', ',
                    self.format_publisher_address(e),
                    ' (',field('year'),')']
                ],
            optional[ sentence [ self.format_isbn(e) ] ],
            optional [ self.format_eprint(e) ],
        ]
        return template
    
    def get_inproceedings_template(self, e):
        template = toplevel [
            sentence [self.format_names('author')],
            self.format_title(e, 'title'),
            words [
                'In',
                sentence [
                    optional[ self.format_editor(e, as_sentence=False) ],
                    self.format_btitle(e, 'booktitle', as_sentence=False),
                    optional [
                        href [
                            field('adsurl',raw=True),
                            self.format_volume_and_series(e, as_sentence=False)
                        ]
                        if 'adsurl' in e.fields and 'doi' in e.fields
                        else
                        self.format_volume_and_series(e, as_sentence=False)
                    ],
                    join [
                        optional[ self.format_pages(e) ],
                        ' (',field('year'),')'
                    ]
                ],
            ]
        ]
        return template
    
    def get_phdthesis_template(self, e):
        template = toplevel [
            sentence [self.format_names('author')],
            self.format_btitle(e, 'title'),
            sentence[
                first_of [
                    optional_field('type'),
                    'PhD thesis',
                ],
                join [
                    field('school'),
                    ', ' if 'address' in e.fields else '',
                    optional_field('address'),
                    ' (',field('year'),')',
                ]
            ]
        ]
        return template    

def register():
    logger.info("Registering astro-style pybtex formatting...")
    register_plugin('pybtex.style.formatting', 'astrostyle', AstroStyle)

def register_with_config(app,config):
    logger.info("Registering astro-style pybtex formatting...")
    if not config.astrorefs_arxiv_url.lower() == 'arxiv.org':
        logger.info(f"Using arxiv URL {config.astrorefs_arxiv_url}")
    global _arxiv_url
    _arxiv_url= config.astrorefs_arxiv_url
    register_plugin('pybtex.style.formatting', 'astrostyle', AstroStyle)



# Quick script to replace all AAS macros in a bibtex file
import os.path
from sphinx.errors import ExtensionError

aas_macros_dict= {
    '\\apjsupp' : 'Astrophys. J. Supp.',
    '\\apjs'    : 'Astrophys. J. Supp.',
    '\\appjlett': 'Astrophys. J. Lett.',
    '\\appjl'   : 'Astrophys. J. Lett.',
    '\\apj'     : 'Astrophys. J.',
    '\\aj'      : 'Astron. J.',
    '\\mnras'   : 'Mon. Not. Roy. Astron. Soc.',
    '\\baas'    : 'Bull. AAS',
    '\\bain'    : 'Bull. Astron. Inst. Netherlands',
    '\\aapr'    : 'Astron. & Astrophys. Rev.',
    '\\aaps'    : 'Astron. & Astrophys. Supp.',
    '\\astap'   : 'Astron. & Astrophys.',
    '\\aap'     : 'Astron. & Astrophys.',
    '\\araa'    : 'Ann. Rev. Astron. Astrophys.',
    '\\actaa'   : 'Acta Astronomica',
    '\\apss'    : 'Astrophys. & Space Sci.',
    '\\jcap'    : 'J. Cosmo & Astropart. Phys.',
    '\\nat'     : 'Nature',
    '\\nar'     : 'New Astron. Rev.',
    '\\na'      : 'New Astron.',
    '\\pra'     : 'Phys. Rev. A',
    '\\prb'     : 'Phys. Rev. B',
    '\\prc'     : 'Phys. Rev. C',
    '\\prd'     : 'Phys. Rev. D',
    '\\pre'     : 'Phys. Rev. E',
    '\\prl'     : 'Phys. Rev. Lett.',
    '\\pasa'    : 'Pub. Astron. Soc. Aus.',
    '\\pasp'    : 'Pub. Astron. Soc. Pac.',
    '\\pasj'    : 'Pub. Astron. Soc. Japan',
    '\\rmxaa'   : 'Rev. Mex. Astron. & Astrofys.',
    '\\ssr'     : 'Space Sci. Rev.',
    '\\applopt' : 'Appl. Opt.',
    '\\ao'      : 'Appl. Opt.',
    '\\azh'     : 'Astron. Zhu.',
    '\\bac'     : 'Bull. Astron. Czech.',
    '\\caa'     : 'Chin. Astron. Astrophys.',
    '\\cjaa'    : 'Chin. J. Astron. Astrophys.',
    '\\icarus'  : 'Icarus',
    '\\jrasc'   : 'J. RAS Can.',
    '\\memras'  : 'Mem. RAS',
    '\\qjras'   : 'Quat. J. RAS',
    '\\skytel'  : 'Sky & Telescope',
    '\\solphys' : 'Sol. Phys.',
    '\\sovast'  : 'Sov. Astron.',
    '\\zap'     : 'ZeitSch. Astrophys.',
    '\\iaucirc' : 'IAU Circs.',
    '\\aplett'  : 'Astrophys. Lett.',
    '\\apspr'   : 'Astrophys. Space Phys. Res.',
    '\\fcp'     : 'Fund. Cosm. Phys.',
    '\\gca'     : 'Geochim. Cosmochim. Acta',
    '\\grl'     : 'Geophys. Res. Lett',
    '\\jcp'     : 'J. Chem. Phys.',
    '\\jgr'     : 'J. Geophys. Res.',
    '\\jqsrt'   : 'J. Quant. Spec. Rad. Trans.',
    '\\memsai'  : 'Mem. Soc. Astron. Ital.',
    '\\nphysa'  : 'Nucl. Phys. A',
    '\\physrep' : 'Phys. Rep.',
    '\\physscr' : 'Phys. Scrip.',
    '\\planss'  : 'Plan. Space. Sci.',
    '\\procspie': 'Proc. SPIE'
    }

def resolve(app):
    if not app.config.astrorefs_resolve_aas_macros:
        return
    if app.config.astrorefs_resolve_aas_macros_infile is None \
       or app.config.astrorefs_resolve_aas_macros_outfile is None:
        raise ExtensionError('sphinx-astrorefs: when resolving AAS macros, need to give original and target bib file name as "astrorefs_resolve_aas_macros_infile" and "astrorefs_resolve_aas_macros_outfile"')
    with open(os.path.join(app.env.srcdir,
                        app.config.astrorefs_resolve_aas_macros_infile),'r') \
                      as infile:
        with open(os.path.join(app.env.srcdir,
                        app.config.astrorefs_resolve_aas_macros_outfile),'w') \
                        as outfile:
            for line in infile:
                for key in aas_macros_dict.keys():
                    line= line.replace(key,aas_macros_dict[key])
                outfile.write(line)
    # Re-do this initialization to make sure the bibtex file is found
    if hasattr(app.config,'bibtex_bibfiles'):
        app.env.get_domain('cite').__init__(app.env)



# Test that AAS macros are correctly resolved (bc a little tricky sometimes...)
import os, os.path
import tempfile
import filecmp
import pytest
from sphinx.errors import ExtensionError
from sphinx_astrorefs import resolve_aas

# Mock up classes that act like Sphinx' app and env used in resolve_aas
# (because I can't easily figure out how to use Sphinx's testing features)

class MockSphinxConfig:
    def __init__(self,
                 astrorefs_resolve_aas_macros=False,
                 astrorefs_resolve_aas_macros_infile=None,
                 astrorefs_resolve_aas_macros_outfile=None):
        self.astrorefs_resolve_aas_macros= \
            astrorefs_resolve_aas_macros
        self.astrorefs_resolve_aas_macros_infile= \
            astrorefs_resolve_aas_macros_infile
        self.astrorefs_resolve_aas_macros_outfile= \
            astrorefs_resolve_aas_macros_outfile

class MockSphinxEnv:
    def __init__(self,srcdir=None):
        self.srcdir= srcdir

class MockSphinxApp:
    def __init__(self,
                 astrorefs_resolve_aas_macros=False,
                 astrorefs_resolve_aas_macros_infile=None,
                 astrorefs_resolve_aas_macros_outfile=None):
        self.config= MockSphinxConfig(\
     astrorefs_resolve_aas_macros=astrorefs_resolve_aas_macros,
     astrorefs_resolve_aas_macros_infile=astrorefs_resolve_aas_macros_infile,
     astrorefs_resolve_aas_macros_outfile=astrorefs_resolve_aas_macros_outfile)
        self.env= MockSphinxEnv(srcdir='./')
        
def test_resolve_aas():
    # Generate file to be resolved and the truth file
    file, resolve_file= tempfile.mkstemp(dir=os.getcwd())
    os.close(file) #Easier this way
    file, resolved_file= tempfile.mkstemp(dir=os.getcwd())
    os.close(file) #Easier this way
    file, truth_file= tempfile.mkstemp(dir=os.getcwd())
    os.close(file) #Easier this way
    try:
        with open(resolve_file,'w') as resolve:
            with open(truth_file,'w') as truth:
                for key in resolve_aas.aas_macros_dict:
                    resolve.write('{}\n'.format(key))
                    truth.write('{}\n'.format(resolve_aas.aas_macros_dict[key]))
        # Now resolve macros and check against truth
        app= MockSphinxApp(astrorefs_resolve_aas_macros=True,
                           astrorefs_resolve_aas_macros_infile=resolve_file,
                           astrorefs_resolve_aas_macros_outfile=resolved_file)
        resolve_aas.resolve(app)
        # Now check resolved against truth
        assert filecmp.cmp(resolved_file,truth_file)
    except:
        raise
    finally:
        if os.path.exists(resolve_file):
            os.remove(resolve_file)
        if os.path.exists(resolved_file):
            os.remove(resolved_file)
        if os.path.exists(truth_file):
            os.remove(truth_file)
    return None

# Test that nothing happens when astrorefs_resolve_aas_macros=False
def test_resolve_aas_false():
    # Generate file to be resolved and the truth file
    file, resolve_file= tempfile.mkstemp(dir=os.getcwd())
    os.close(file) #Easier this way
    file, resolved_file= tempfile.mkstemp(dir=os.getcwd())
    os.close(file) #Easier this way
    # Remove resolved so it doesn't exist and shouldn't get created
    os.remove(resolved_file)
    try:
        with open(resolve_file,'w') as resolve:
            for key in resolve_aas.aas_macros_dict:
                resolve.write('{}\n'.format(key))
        # Now don't do anything
        app= MockSphinxApp(astrorefs_resolve_aas_macros=False,
                           astrorefs_resolve_aas_macros_infile=resolve_file,
                           astrorefs_resolve_aas_macros_outfile=resolved_file)
        resolve_aas.resolve(app)
        assert not os.path.exists(resolved_file)
    except:
        raise
    finally:
        if os.path.exists(resolve_file):
            os.remove(resolve_file)
        if os.path.exists(resolved_file):
            os.remove(resolved_file)
    return None

# Test that one gets an error when not specifying filenames
def test_resolve_aas_error():
    app= MockSphinxApp(astrorefs_resolve_aas_macros=True,
                       astrorefs_resolve_aas_macros_infile=None,
                       astrorefs_resolve_aas_macros_outfile=None)
    with pytest.raises(ExtensionError):
        resolve_aas.resolve(app)
    return None
